import itertools
from abc import ABC, abstractmethod
from argparse import Namespace
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

from dbt.adapters.factory import get_adapter, register_adapter
from dbt.clients.system import run_cmd
from dbt.config import PROFILES_DIR, RuntimeConfig
from dbt.contracts.graph.manifest import Manifest
from dbt.exceptions import RuntimeException
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.perf_utils import get_full_manifest

import attr


@attr.s(auto_attribs=True)
class DOBTBaseTask(ABC):
    args: Namespace = attr.ib(
        default=Namespace(profiles_dir=PROFILES_DIR, project=None, profile=None)
    )
    config: Optional[RuntimeConfig] = attr.ib(default=None, init=False)
    manifest: Optional[Manifest] = attr.ib(default=None, init=False)
    adapter: Optional[Any] = attr.ib(default=None, init=False)

    @abstractmethod
    def run(self):
        """Run Dobt task."""

    @staticmethod
    def slack_log(text: str, log=True):
        section = {"type": "section", "text": {"type": "mrkdwn", "text": text}}
        if log:
            logger.info(text)
        return section

    @staticmethod
    def file_to_model(filename: str) -> str:
        """Convert file uri to model name.

        Args:
            filename (str): filename uri.

        Returns:
            str: model name
        """
        return Path(filename).stem

    @staticmethod
    def git_command(cmd: List[str], filter_suffix: str = ".sql"):
        """Get current modified and untracked models from git.

        Args:
            cmd (List[str]): command to run
            filter_suffix (str): filter changes by suffix. Default .sql.

        Raises:
            RuntimeException: no models are found.

        Returns:
            Set[str]: Set of modifed and untracked files.
        """
        logger.debug(f"Running git command: {' '.join(cmd)}")
        out, _ = run_cmd(".", cmd, env={"LC_ALL": "C"})
        changes = out.decode("utf-8").strip().split("\n")
        result = {
            DOBTBaseTask.file_to_model(change)
            for change in changes
            if Path(change).suffix == filter_suffix
        }
        if not result:
            raise RuntimeException(f"No model changes found. Cmd: {' '.join(cmd)}")
        return result

    def _get_config(self):
        self.config = RuntimeConfig.from_args(self.args)
        register_adapter(self.config)

    def _get_adapter(self):
        self._get_config()
        logger.debug("Getting adapter.")
        self.adapter = get_adapter(self.config)

    def _load_manifest(self):
        """Load DBT manifest."""
        self._get_config()
        logger.debug("Loading manifest.")
        self.manifest = get_full_manifest(self.config)

    def _get_enabled_models(
        self,
        filter_fce=lambda x: str(x[1].resource_type) == "model" and x[1].config.enabled,
    ) -> Dict[Any, Any]:
        """Set enabled models from manifest.

        Args:
            filter_fce ([type], optional): Function to filter models.
                Defaults to lambdax:x[1].configandx[1].config.enabled.

        Returns:
            Dict[Any]: Dictionary of models.
        """
        if not self.manifest:
            self._load_manifest()
        if self.manifest:
            return dict(filter(filter_fce, self.manifest.nodes.items()))
        else:
            return {}


@attr.s(auto_attribs=True)
class DOBTDBSchemaTask(DOBTBaseTask):
    database: List[str] = attr.ib(default=attr.Factory(list))
    schema: List[str] = attr.ib(default=attr.Factory(list))

    db_schemas: Set[Tuple[str, str]] = attr.ib(default=attr.Factory(set), init=False)

    def _get_db_schema_manifest(self):
        logger.info("Loading list of databases and schemas from dbt manifest...")
        self._load_manifest()
        db_schemas = set()
        for k, v in self.manifest.nodes.items():
            db_schemas.add((v.database, v.schema))
        if db_schemas:
            return db_schemas
        else:
            raise RuntimeException(
                f"No schemas found in your dbt project. "
                f"Are you sure it is initialized?"
            )

    def _set_db_schema_combination(self):
        # If only schema is specified, take database from current dbt profile is taken
        if self.schema and not self.database:
            db = self.config.credentials.database
            db_schemas = set()
            for s in self.schema:
                quote = self.adapter._make_match_kwargs(db, s, "")
                if self.adapter.check_schema_exists(quote["database"], quote["schema"]):
                    db_schemas.add((quote["database"], quote["schema"]))
                else:
                    logger.info(
                        f"Schema {quote['schema']} not found in database "
                        f"{quote['database']}. Skipping"
                    )
            if db_schemas:
                self.db_schemas.update(db_schemas)
            else:
                raise RuntimeException(
                    f"No specified schemas was found in database {quote['database']}."
                )

        # If only database is specified, take all schemas from database
        # and then from manifest
        elif self.database and not self.schema:
            logger.info("Loading list of schemas from database...")
            for db in self.database:
                quote = self.adapter._make_match_kwargs(db, "", "")
                db_schemas = set()
                try:
                    tmp_schemas = self.adapter.list_schemas(quote["database"])

                    for s in tmp_schemas:
                        if s.lower() not in ["information_schema", "pg_catalog"]:
                            quote = self.adapter._make_match_kwargs(db, s, "")
                            db_schemas.add((quote["database"], quote["schema"]))
                except RuntimeException:
                    logger.info(f"Database {db} not found. Skipping.")
            db_schemas.update(self._get_db_schema_manifest())
            self.db_schemas.update(db_schemas)

        # If database and schema
        elif self.database and self.schema:
            logger.info("Checking list of schemas in database...")
            db_schema_product = list(itertools.product(self.database, self.schema))
            db_schemas = set()
            for db_schema in db_schema_product:
                quote = self.adapter._make_match_kwargs(db_schema[0], db_schema[1], "")
                try:
                    self.adapter.check_schema_exists(quote["database"], quote["schema"])
                    db_schemas.add((quote["database"], quote["schema"]))
                except RuntimeException:
                    logger.info(
                        f"Schema {quote['schema']} was not found in database "
                        f"{quote['database']}. Skipping"
                    )
            if db_schemas:
                self.db_schemas.update(db_schemas)
            else:
                raise RuntimeException(
                    f"No specified schemas and databases were found."
                )

        # Otherwise take database and schema from manifest
        else:
            self.db_schemas = self._get_db_schema_manifest()
