import itertools
from typing import Any, Dict, List, Optional, Set, Tuple

from dbt.exceptions import RuntimeException
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.node_types import NodeType
from dbt.ui.printer import green

import attr
from dobt.task.base import DOBTDBSchemaTask


@attr.s(auto_attribs=True)
class CompareTask(DOBTDBSchemaTask):
    models_manifest: Dict[Tuple[str, str, str], Any] = attr.ib(
        default=attr.Factory(dict), init=False
    )
    rels_db: Dict[Tuple[str, str, str], Any] = attr.ib(
        default=attr.Factory(dict), init=False
    )

    @classmethod
    def from_args(cls, args):
        return cls(database=args.database, schema=args.schema, args=args)

    def _set_models_from_manifest(self):
        logger.info("Getting list of models...")
        self._load_manifest()
        result = {}
        for k, v in self.manifest.nodes.items():
            if (
                v.config
                and v.resource_type in NodeType.refable()
                and v.config.materialized != "ephemeral"
                and (v.database, v.schema) in self.db_schemas
            ):
                quote = self.adapter._make_match_kwargs(v.database, v.schema, v.alias)
                result[quote["database"], quote["schema"], quote["identifier"]] = v

        self.models_manifest = result

    def _set_rels_from_db(self):
        logger.info("Getting list of relations...")
        result = {}
        for db, schema in self.db_schemas:
            rels_tmp = {
                (rel.database, rel.schema, rel.name): rel
                for rel in self.adapter.list_relations(db, schema)
            }
            result.update(rels_tmp)

        self.rels_db = result

    def compare(self) -> Tuple[Set[Any], Set[Any]]:
        """Compare relations in database and models in dbt manifest and returns diff.

        Returns:
            Tuple[Set[str], Set[str]]: Tuple of diffences between db and dbt manifest.
                First value contains db only relations and second dbt only models.
        """
        self._get_adapter()
        with self.adapter.connection_named("dobt_compare"):
            self._set_db_schema_combination()
            self._set_rels_from_db()
        self._set_models_from_manifest()

        rels_keys = set(self.rels_db.keys())
        models_keys = set(self.models_manifest.keys())

        db_only = rels_keys - models_keys
        model_only = models_keys - rels_keys

        return db_only, model_only

    def _interpret(self, db_only, model_only):
        results = []
        success = True

        if not model_only and not db_only:
            logger.info(
                green(
                    f"All clear. Thera are no discrepancies "
                    f"between database and dbt models."
                )
            )

        if db_only:
            logger.info("=" * 70)
            results.append(
                self.slack_log("Following relations do not match any model:")
            )
            logger.info("=" * 70 + "\n")
            tmp = {}
            for db, schema, rel_name in db_only:
                db_object = self.rels_db.get((db, schema, rel_name))
                tmp.setdefault((db, schema), []).append(
                    f"{rel_name} [{db_object.type}]"
                )
            for key, value in tmp.items():
                msg = f"{key[0]}.{key[1]}" + "\n• " + "\n• ".join(value) + "\n"
                results.append(self.slack_log(msg))
            success = False

        if model_only:
            logger.info("=" * 70)
            results.append(
                self.slack_log("Following models are not materialized in database:")
            )
            logger.info("=" * 70 + "\n")
            tmp = {}
            for db, schema, rel_name in model_only:
                model = self.models_manifest.get((db, schema, rel_name))
                tmp.setdefault((db, schema), []).append(
                    f"{model.unique_id} [{model.config.materialized}] "
                    f"({model.original_file_path})"
                )
            for key, value in tmp.items():
                msg = f"{key[0]}.{key[1]}" + "\n• " + "\n• ".join(value) + "\n"
                results.append(self.slack_log(msg))
            success = False

        return results, success

    def run(self):
        db_only, model_only = self.compare()
        slack_msg, success = self._interpret(db_only, model_only)
        results = {
            "results": {"db_only": list(db_only), "model_only": list(model_only)},
            "slack_msg": slack_msg,
        }

        return results, success
