import inspect
import json
from datetime import datetime
from typing import Dict, List, Optional, Set

import dbt.main as dbt
import dbt.task.run as run_task
from dbt.exceptions import RuntimeException
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.logger import log_manager
from dbt.ui.printer import green

import attr
from dateutil.parser import parse
from dobt.constants import (
    GIT_CURRENT_CHANGES,
    GRANULARITY_STEP,
    START_MESSAGE,
    get_commit_diff,
)
from dobt.task.base import DOBTBaseTask


@attr.s(auto_attribs=True)
class RunTask(DOBTBaseTask):
    start: str = datetime.now().strftime("%Y-%m-%d")
    end: str = datetime.now().strftime("%Y-%m-%d")
    granularity: str = "days"
    increment: int = 1
    mater_type: List[str] = attr.Factory(list)
    cached_changes: bool = False
    commit_diff: bool = False
    commit_id: str = "HEAD^"
    models: Optional[List[str]] = None

    @classmethod
    def from_args(cls, args):
        return cls(
            start=args.start,
            end=args.end,
            granularity=args.granularity,
            increment=args.increment,
            mater_type=args.mater_type,
            cached_changes=args.cached_changes,
            commit_diff=args.commit_diff,
            commit_id=args.commit_id,
            models=args.models,
            args=args,
        )

    @staticmethod
    def _items_intersection(*items: Set[str]) -> Optional[List[str]]:
        """Find intersection between changes.

        Args:
            items (Dict[str, Set): dictionary of items to find intersection.

        Raises:
            RuntimeException: no models are found

        Returns:
            Optional[Set[str]]: intersected items
        """
        # filter out keys without values
        filtered = [v for v in items if v]
        result: Optional[List[str]] = None
        if filtered:
            # result models are all that intersect
            tmp_result = set.intersection(*filtered)
            if not tmp_result:
                raise RuntimeException("No models match your arguments.")
            result = list(tmp_result)
        return result

    @staticmethod
    def _timestamp_vars(timestamp: datetime, prefix: Optional[str]) -> Dict[str, str]:
        """Get dictionary of formatted useful time variables. Prefix in case you need multiple
        timestamp instances like start, next, end.

        Args:
            timestamp (datetime): datetime to be formated
            prefix (Optional[str]): prefix for dictionary

        Returns:
            Dict[str, str]: dictionary of useful formatted variables
        """
        prefix = f"{prefix}_" if prefix else ""
        return {
            f"{prefix}timestamp": timestamp.strftime("%Y-%m-%d %H:%M:%S"),
            f"{prefix}date": timestamp.strftime("%Y-%m-%d"),
            f"{prefix}time": timestamp.strftime("%H:%M:%S"),
            f"{prefix}year": timestamp.strftime("%Y"),
            f"{prefix}month": timestamp.strftime("%m"),
            f"{prefix}week": timestamp.strftime("%W"),
            f"{prefix}day": timestamp.strftime("%d"),
            f"{prefix}daynumber": timestamp.strftime("%j"),
            f"{prefix}hour": timestamp.strftime("%H"),
            f"{prefix}minute": timestamp.strftime("%M"),
            f"{prefix}second": timestamp.strftime("%S"),
        }

    def models_by_type(self) -> Set[str]:
        """Get set of models from manifest filtered by model type (view, table, incremental).

        Raises:
            RuntimeException: no models are found

        Returns:
            Set[str]: set of filtered models.
        """
        result: Set[str] = set()
        enabled_models = self._get_enabled_models()
        for key, value in enabled_models.items():
            if value.config.materialized in self.mater_type:
                result.add(value.name)
        if not result:
            raise RuntimeException(
                f"No models match materializations: {self.mater_type}"
            )
        return result

    def time_range(self) -> List[Dict[str, str]]:
        """Get list of intevals and date variables for running the command.

        Returns:
            List[Dict[str, str]]: List of intervals to be run
        """
        start_timestamp = parse(self.start)
        end_timestamp = parse(self.end)
        time_range = []

        while start_timestamp <= end_timestamp:
            next_timestamp = start_timestamp + GRANULARITY_STEP[self.granularity](
                self.increment
            )
            time_vars = {
                "date_valid": start_timestamp.strftime("%Y-%m-%d"),
                "granularity": self.granularity,
                "increment": str(self.increment),
                **self._timestamp_vars(start_timestamp, "start"),
                **self._timestamp_vars(next_timestamp, "next"),
                **self._timestamp_vars(end_timestamp, "end"),
            }
            time_range.append(time_vars)
            start_timestamp = next_timestamp

        return time_range

    def _update_models(self):
        git_changes = set()
        mater_changes = set()
        user_input = set(self.models) if self.models else None

        # get git changes
        if self.cached_changes:
            git_changes.update(self.git_command(GIT_CURRENT_CHANGES))
        if self.commit_diff:
            cmd = get_commit_diff(self.commit_id)
            git_changes.update(self.git_command(cmd))

        # Get models by mater type
        if self.mater_type:
            mater_changes.update(self.models_by_type())

        # Find intersection between models
        models = self._items_intersection(git_changes, mater_changes, user_input)
        return models

    def _set_vars(self, time_interval: Dict[str, str]):
        if hasattr(self.args, "vars"):
            input_vars = json.loads(self.args.vars)
            time_interval.update(input_vars)
        self.args.vars = json.dumps(time_interval)

    def run(self):
        # Set models to selected models and cls back to dbt RunTask
        self.args.models = self._update_models()
        self.args.cls = run_task.RunTask

        time_range = self.time_range()
        i = 1
        msg = inspect.cleandoc(START_MESSAGE)
        exec_results = []
        exec_success = True
        for time_interval in time_range:
            self._set_vars(time_interval)
            msg = msg.format(
                i=i,
                iter_no=len(time_range),
                start=self.start,
                end=self.end,
                granularity=self.granularity,
                increment=self.increment,
                start_timestamp=time_interval.get("start_timestamp"),
                next_timestamp=time_interval.get("next_timestamp"),
            )
            logger.info(green(msg))
            task, res = dbt.run_from_args(self.args)
            exec_results.append(res)
            success = task.interpret_results(res)
            exec_success = all([success, exec_success])
            log_manager.reset_handlers()
            i += 1

        return exec_results, exec_success
